/***************************************************************************\
  "parser.h" - Parser Module (Phase C)
  Revision 21; 10 Nov 2002 - 21 Feb 2004
  Copyright (C) 2002-2004  Kevin P. Albrecht
\***************************************************************************/

#ifndef PARSER_H
#define PARSER_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //

// Local //
#include "global.h"
#include "error.h"
#include "icm.h"
#include "k_string.h"
#include "table.h"
#include "token.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// Internal //
Token *mToken;    // current token
Table *mTable;    // symbol table
TokenCode mCode;  // code of current token
Symbol **mSymbolList;
int mSymbolListCurr;

// External //
void parser_init ();
void parser_deinit ();
void parser_parse ();

// Internal - Parse Statements //
void mParseStatement ();
void mParseDeclaration ();
void mParseLabel ();
void mParseZeroParamOps ();
void mParseJumpOps ();
void mParsePopOps ();
void mParsePushOps ();
void mParseDebugOps ();
void mParseFileIOOps ();

// Internal - Parse Expressions //
Symbol *mParseExpression ();

// Internal - Special //
void mSkipEOL ();
void mEnsureSkipEOL ();
bool mSync (TokenCode aCode, char *aError);
void mSeekToEOL ();
void mGetToken ();
Symbol *mTableEnter (char *key);

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// parser_init
//// Initialize parser module
void parser_init ()
{
  int i;
  mSymbolListCurr = 0;

  // Allocate space for the symbol list, used to free memory at end of program //
  mSymbolList = (Symbol **) KALLOC(sizeof (Symbol *) * GLOBAL_MAX_SYMBOLS);
  for (i = 0; i < GLOBAL_MAX_SYMBOLS; i++)
  {
    mSymbolList[i] = NULL;
  }  

}

//// parser_deinit
//// Deinitialize parser module
void parser_deinit ()
{
  int i;

  for (i = 0; mSymbolList[i] != NULL; i++)
  {
    sym_del (mSymbolList[i]);
  }

  KFREE(mSymbolList);
}

//// parser_parse
void parser_parse ()
{
  /////////////
  // PARSING //
  /////////////

  mTable = tab_new ();
  mToken = tok_new ();      //
  mToken->code = ttStart; // Passed to lex_get at the start of the file
  mGetToken();              //
  mSkipEOL();

  do {
    // Parse a Statement //
    mParseStatement ();

    // Handles empty files //
    if (mCode == ttEOF) break;

    // check for EOL //
    mEnsureSkipEOL();

  } while (mCode != ttEOF);

  KFREE(mToken);
  tab_del (mTable);
}

///////////////////
// Special Parse //
///////////////////

//// mParseStatement
void mParseStatement ()
{
  switch (mCode)
  {
    case (ttDB):
    case (ttDI):
    case (ttDR):
    case (ttDS):
      mParseDeclaration ();
      break;
    case (ttIdentifier):
      mParseLabel ();
      break;
    case (ttEQ):
    case (ttLT):
    case (ttGT):
    case (ttOR):
    case (ttNOP):
    case (ttRET):
    case (ttTOI):
    case (ttTOS):
    case (ttTOR):
    case (ttTOB):
    case (ttADDI):
    case (ttSUBI):
    case (ttMULI):
    case (ttDIVI):
    case (ttPOWI):
    case (ttMODI):
    case (ttADDR):
    case (ttSUBR):
    case (ttMULR):
    case (ttDIVR):
    case (ttPOWR):
    case (ttMODR):
    case (ttNEQ):
    case (ttAND):
    case (ttCAT):
    case (ttCIN):
    case (ttASC):
    case (ttSTOP):
    case (ttCOPY):
    case (ttLTEQ):
    case (ttGTEQ):
    case (ttCOUT):
    case (ttRAND):
    case (ttSWAP):
    case (ttNOT):
    case (ttNEG):
    case (ttDBGON):
      mParseZeroParamOps ();
      break;
    case (ttJT):
    case (ttJF):
    case (ttJMP):
    case (ttCALL):
      mParseJumpOps ();
      break;
    case (ttPOP):
    case (ttPOPL):
    case (ttPOPP):
      mParsePopOps ();
      break;
    case (ttPUSH):
    case (ttPUSHL):
    case (ttPUSHP):
      mParsePushOps ();
      break;
    case (ttDBGLINE):
    case (ttDBGFILE):
      mParseDebugOps ();
      break;
    case (ttFIO_OR):
    case (ttFIO_OW):
    case (ttFIO_C):
    case (ttFIO_R):
    case (ttFIO_W):
    case (ttFIO_EOF):
      mParseFileIOOps ();
      break;
    case (ttEOL):
    case (ttEOF):
      break;
    default:
      err (true, EF_INVALID_PARSE_POINT, gvsTokenCodeStrings[mCode]);
      erre (EFF_TERMINATE);
      break;
  }
}

///////////////////////
// Statement Parsing //
///////////////////////

void mParseDeclaration ()
{
  Type type;
  Symbol *symbol;

  switch (mCode)
  {
    case (ttDB):
      type = tBool;
      break;
    case (ttDI):
      type = tInt;
      break;
    case (ttDR):
      type = tRat;
      break;
    case (ttDS):
      type = tStr;
      break;
  }

  mGetToken (); // now: identifier
  mSync (ttIdentifier, EF_UNEXPECTED_TOKEN);

  do
  {
    // Search for the variable's identifier and enter it //
    symbol = tab_search (mTable, mToken->lexeme);
    if (symbol)
    {
      // The symbol has already been defined, cause error //
      err (true, EF_ALREADY_DECLARED, mToken->lexeme);
      erre (EFF_RECOVER);
    }
    else
    {
      // The symbol is created //
      symbol = mTableEnter(mToken->lexeme);
      sym_set (symbol, repIdent, type, NULL);
    }

    mGetToken (); // now: identifier or EOL
  } while (mCode != ttEOL);
}

//// mParseLabel
void mParseLabel ()
{
  Symbol *symbol;

  // Search for the label's identifier and enter it //
  symbol = tab_search (mTable, mToken->lexeme);
  if (symbol) // label must have been created by a jump op
  {
    if (symbol->type == tLabel)
    {
      KFREE(symbol->value);
      symbol->value = k_new_n (gxCurr);
      symbol->special = true; // means that label has been defined
    }
    else
    {
      err (true, EF_TYPE_INCOMPATIBLE, gvsTypeStrings[symbol->type]);
      erre (EFF_RECOVER);
    }
  }
  else if (!symbol) // create label
  {
    symbol = mTableEnter (mToken->lexeme);
    sym_set (symbol, repLiteral, tLabel, &gxCurr);
    symbol->special = true; // means that label has been defined
  }

  mGetToken (); // now: EOL or statement

  // The label is directly followed by a statement //
  if (mCode != ttEOL)
  {
    mParseStatement ();
  }
}

//// mParseZeroParamOps
void mParseZeroParamOps ()
{
  icm_put (statement_new (mToken->code, NULL));
  mGetToken (); // now: EOL
}

//// mParseJumpOps
void mParseJumpOps ()
{
  Symbol *symbol;

  TokenCode tc = mToken->code;

  mGetToken (); // now: identifier

  // Search for the label's identifier and enter it //
  symbol = tab_search (mTable, mToken->lexeme);
  if (symbol) // label must have been created already
  {
    if (symbol->type == tLabel)
    {
      icm_put (statement_new (tc, symbol));
    }
    else
    {
      err (true, EF_TYPE_INCOMPATIBLE, gvsTypeStrings[symbol->type]);
      erre (EFF_RECOVER);
    }
  }
  else // create label
  {
    symbol = mTableEnter (mToken->lexeme);
    sym_set (symbol, repLiteral, tLabel, NULL);
    symbol->special = false; // means that label is not yet defined;

    icm_put (statement_new (tc, symbol));
  }

  mGetToken (); // now: EOL
}

//// mParsePopOps
void mParsePopOps ()
{
  Symbol *symbol;

  mGetToken (); // now: variable

  mSync (ttIdentifier, EF_UNEXPECTED_TOKEN);

  // Get token's symbol //
  symbol = tab_search (mTable, mToken->lexeme);
  if (!symbol)
  {
    // The symbol has not been declared, cause error //
    err (true, EF_UNDECLARED_IDENTIFIER, mToken->lexeme);
    erre (EFF_RECOVER);

    // Enter the symbol as str for recovery //
    symbol = mTableEnter (mToken->lexeme);
    sym_set (symbol, repIdent, tStr, NULL);
  }

  icm_put (statement_new (ttPOP, symbol));

  mGetToken (); // now: EOL
}

//// mParsePushOps
void mParsePushOps ()
{
  Symbol *symbol;

  mGetToken (); // now first token of expression

  symbol = mParseExpression ();

  icm_put (statement_new (ttPUSH, symbol));
}

//// mParseDebugOps
void mParseDebugOps ()
{
  Symbol *symbol;
  TokenCode code;

  code = mCode;

  mGetToken (); // now first token of expression

  symbol = mParseExpression ();

  if (code == ttDBGFILE)
  {
    if (symbol->type != tStr)
    {
      err (true, EF_TYPE_INCOMPATIBLE, "expected string");
      erre (EFF_RECOVER);
    }
  }
  else
  {
    if (symbol->type != tInt)
    {
      err (true, EF_TYPE_INCOMPATIBLE, "expected integer");
      erre (EFF_RECOVER);
    }
  }

  icm_put (statement_new (code, symbol));
}

void mParseFileIOOps ()
{
  Symbol *symbol;
  TokenCode opCode = mCode;

  switch (opCode)
  {
    case (ttFIO_OR):
    case (ttFIO_OW):
    {
      mGetToken (); // now: expression

      symbol = mParseExpression ();

      if (symbol->type != tStr)
      {
        err (true, EF_TYPE_INCOMPATIBLE, "expected string");
        erre (EFF_RECOVER);
      }

      icm_put (statement_new (opCode, symbol));
      break;
    }
    case (ttFIO_C):
    case (ttFIO_R):
    case (ttFIO_W):
    case (ttFIO_EOF):
    {
      mGetToken (); // now: expression

      symbol = mParseExpression ();

      if (symbol->type != tInt)
      {
        err (true, EF_TYPE_INCOMPATIBLE, "expected integer");
        erre (EFF_RECOVER);
      }

      icm_put (statement_new (opCode, symbol));
      break;
    }
  }
}

////////////////////////
// Expression Parsing //
////////////////////////

//// mParseExpression
Symbol *mParseExpression ()
{
  Symbol *symbol;
  Type type;

  switch (mCode)
  {
    case ttIdentifier:
    {
      // Search for the identifier //
      symbol = tab_search (mTable, mToken->lexeme);
      if (!symbol)
      {
        // Not found; Undefined identifier error; Enter it in symbol table //
        err (true, EF_UNDECLARED_IDENTIFIER, mToken->lexeme);
        erre (EFF_RECOVER);

        symbol = mTableEnter (mToken->lexeme);
      }

      mGetToken (); // now: EOL

      return symbol;
      break;
    }
    case ttIntLiteral:
    case ttBoolLiteral:
    case ttRatLiteral:
    case ttStrLiteral:
    {
      if (mCode == ttIntLiteral) type = tInt;
      else if (mCode == ttBoolLiteral) type = tBool;
      else if (mCode == ttRatLiteral) type = tRat;
      else if (mCode == ttStrLiteral) type = tStr;

      // Search for it and enter it in the symbol table //
      symbol = tab_search (mTable, mToken->lexeme);
      if (!symbol)
      {
        symbol = mTableEnter (mToken->lexeme);
        sym_set (symbol, repLiteral, type, mToken->value);
      }

      mGetToken ();
      return symbol;
      break;
    }
    default:
    {
      err (true, EF_INVALID_EXPRESSION, NULL);
      erre (EFF_RECOVER);

      // BUGFIX (29 July 2002, build 0103):
      //   this is added so that a NULL pointer is not returned
      symbol = mTableEnter ("<invalid expression>");
      sym_set (symbol, repLiteral, tStr, "<invalid expression>");

      return symbol;
      break;
    }
  }
}

//////////////
// Special  //
//////////////

void mSkipEOL ()
{
  // Skip over this EOL and any blank lines before parsing the next statement //
  while (mCode == ttEOL)
  {
    mGetToken ();
  }
}

void mEnsureSkipEOL ()
{
  // Make sure that an end of line has been reached //
  mSync (ttEOL, EF_MISSING_EOL);

  // Skip over this EOL and any blank lines before parsing the next statement //
  while (mCode == ttEOL)
  {
    mGetToken ();
  }
}

//// mSync
//// Make sure that one of the codes has been reached
bool mSync (TokenCode aCode, char *aError)
{
  if (mCode != aCode) // code did not match
  {
    if ((strcmp (mToken->lexeme, "") == 0) || (mToken->lexeme == NULL))
      err (true, aError, gvsTokenCodeStrings[mToken->code]);
    else
      err (true, aError, mToken->lexeme);      
    erre (EFF_RECOVER);
    mSeekToEOL ();
    return false;
  }
  return true; // code did match
}

void mSeekToEOL ()
{
  // To allow the parser to continue, skip tokens until an EOL is reached //
  while ((mCode != ttEOL) && (mCode != ttEOF))
  {
    mGetToken ();
  }
}

void mGetToken ()
{
  mToken = lex_get(mToken);
  mCode = mToken->code;
}

// This is used to add all symbols to a list to be deleted //
//  at program termination; Modules that KALLOC a Symbol   //
//  and KFREE it should not put it on this list.           //
Symbol *mTableEnter (char *key)
{
  Symbol *symbol = tab_enter (mTable, key);

  mSymbolList[mSymbolListCurr] = symbol;
  mSymbolListCurr++;

  return symbol;
}

#endif
